/* merged (for a sufficiently loose meaning of "merge") calls to HDF-EOS
   4/5.  Including this should get you set to write programs that mess
   with HDF-EOS 4/5 files alike.  The limitations I'm willing to live
   with here are:
   - One HDF-EOS 4/5 file open at a time
   - inquiry/read functions mostly Just Work
   - write functions can fail if you try to write H5-only stuff to H4 file
   - we write H5 files by default
*/
#ifndef H45_H
#define H45_H
/* For this library, we need all four sets of headers, and they need
   to go in in this order (because of the multiple definition of
   uintn in HDF4 and HDF_EOS5)
*/

#include "hdf.h"
#include "HdfEosDef.h"
#include "hdf5.h"
#include "HE5_HdfEosDef.h"

#include "trace.h"

/* values to tell which type of file you're processing */
#define H4FILE -1
#define H5FILE 1

int h4h5which(void);
int h4h5choose(int which);

hid_t h4h5_SWopen(const char *filename, uintn access);
hid_t h4h5_SWcreate(hid_t swhandle, const char *swath_name);
long h4h5_SWinqswath(const char* filename, char *swathlist, long *strbufsize);
hid_t h4h5_SWattach(hid_t fid, const char*swathname);
herr_t h4h5_SWclose(hid_t fid);
herr_t h4h5_SWdetach(hid_t swhandle);
long h4h5_SWinqattrs(hid_t swhandle, char *attrlist, long *strbufsize);
long h4h5_SWinqdims(hid_t swhandle, char *dimnamebuf, hsize_t *dimsizebuf);
long h4h5_SWinqmaps(hid_t swhandle, char *dimmap, long *offset, long *increment);
long h4h5_SWnentries(hid_t swhandle, int entrycode, long *strbufsize);
herr_t h4h5_SWreadattr(hid_t swhandle, char *attrname, void *datbuf);
herr_t h4h5_SWreadfield(hid_t swhandle, char *fieldname, const hssize_t *start, const hsize_t *stride, const hsize_t *edge, void *datbuf);
herr_t h4h5_SWattrinfo(hid_t swhandle, const char *attrname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_SWlocattrinfo(hid_t swhandle, const char *fieldname, const char *attrname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_SWgrpattrinfo(hid_t swhandle, const char *attrname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_SWfieldinfo(hid_t swhandle, char *fieldname, int *rank, hsize_t *dims, H5T_class_t *ntype, char *dimlist, char *maxdimlist);
herr_t h4h5_SWinqdatafields(hid_t swhandle, char *fieldlist, int *rank, H5T_class_t *ntype);
herr_t h4h5_SWinqgeofields(hid_t swhandle, char *fieldlist, int *rank, H5T_class_t *ntype);
herr_t h4h5_SWdefcomp(hid_t swhandle, int compcode, int *compparm);
herr_t h4h5_SWdefdatafield(hid_t swhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge);
herr_t h4h5_SWdefdim(hid_t swhandle, char *dimname, hsize_t dimsize);
herr_t h4h5_SWdefdimmap(hid_t swhandle, char *geodim, char *datadim, hsize_t offset, hsize_t increment);
herr_t h4h5_SWdefgeofield(hid_t swhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge);
herr_t h4h5_SWdefidxmap(hid_t swhandle, char *geodim, char *datadim, long index[]);
herr_t h4h5_SWwriteattr(hid_t swhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_SWwritefield(hid_t swhandle, char *fieldname,  const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf);
herr_t h4h5_SWwritelocattr(hid_t swhandle, const char *fieldname, char *attrname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_SWwritegrpattr(hid_t swhandle, char *attrname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_SWsetalias(hid_t swhandle, char *fieldname, const char *aliaslist);

/* and the profile junk */
herr_t h4h5_PRdefine(hid_t swhandle, const char *prname, char *dimlist, char *maxdimlist, hid_t datatype);
herr_t h4h5_PRinfo(hid_t swhandle, const char *prname, int *rank, hsize_t dims[], hsize_t maxdims[], H5T_class_t *classID, char *dimlist, char *maxdimlist);
herr_t h4h5_PRinquire(hid_t swhandle, char *prnames, int *rank, H5T_class_t *classID);
herr_t h4h5_PRread(hid_t swhandle, const char *prname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf);
herr_t h4h5_PRreclaimspace(hid_t swhandle, const char *prname, void *buffer);
herr_t h4h5_PRwrite(hid_t swhandle, const char *prname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], size_t size, void *databuf);

/* The Grid API */
hid_t h4h5_GDopen(const char *filename, uintn access);
hid_t h4h5_GDcreate(hid_t fid, const char *gridname, long xdimsize, long ydimsize, double upleft[], double upright[]);
long h4h5_GDinqgrid(const char* filename, char *gridlist, long *strbufsize);
hid_t h4h5_GDattach(hid_t fid, const char *gridname);
herr_t h4h5_GDattrinfo(hid_t grhandle, const char *attname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_GDlocattrinfo(hid_t grhandle, const char *fieldname, const char *attname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_GDgrpattrinfo(hid_t grhandle, const char *attname, H5T_class_t *ntype, hsize_t *count);
herr_t h4h5_GDclose(hid_t fid);
herr_t h4h5_GDdetach(hid_t grhandle);
herr_t h4h5_GDfieldinfo(hid_t grhandle, char *fieldname, int *rank, hsize_t *dims, H5T_class_t *ntype, char *dimlist, char *maxdimlist);
herr_t h4h5_GDgridinfo(hid_t grhandle, long *xsize, long *ysize, double *upleft, double *lowright);
long h4h5_GDinqattrs(hid_t grhandle, char *attrlist, long *strbufsize);
herr_t h4h5_GDinqfields(hid_t grhandle, char *fieldlist, int *rank, H5T_class_t *ntype);
long h4h5_GDinqdims(hid_t grhandle, char *dimnamebuf, hsize_t *dimsizebuf);
long h4h5_GDnentries(hid_t grhandle, int entrycode, long *strbufsize);
herr_t h4h5_GDprojinfo(hid_t grhandle, int *projcode, int *zonecode, int *spherecode, double *projparm);
herr_t h4h5_GDreadattr(hid_t grhandle, char *attrname, void *datbuf);
herr_t h4h5_GDreadfield(hid_t grhandle, char *fieldname, const hssize_t *start, const hsize_t *stride, const hsize_t *edge, void *datbuf);
herr_t h4h5_GDdefcomp(hid_t grid_handle, int dimname, int dimsize[]);
herr_t h4h5_GDdefcomtile(hid_t grhandle, int compcode, int *compparm, int tilerank, const hsize_t *tiledim);
herr_t h4h5_GDdefdim(hid_t grhandle, char *dimname, hsize_t dim);
herr_t h4h5_GDdeffield(hid_t grhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge);
herr_t h4h5_GDdeforigin(hid_t grhandle, int origincode);
herr_t h4h5_GDdefpixreg(hid_t grhandle, int prcode);
herr_t h4h5_GDdefproj(hid_t grhandle, int projcode, int zonecode, int spherecode, double projparm[]);
herr_t h4h5_GDdeftile(hid_t grhandle, int tilecode, int tilerank, const hsize_t tiledims[]);
herr_t h4h5_GDwriteattr(hid_t grhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_GDwritefield(hid_t grhandle, const char *fieldname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf);
herr_t h4h5_GDwritegrpattr(hid_t grhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_GDwritelocattr(hid_t grhandle, const char *fieldname, const char *attrname, hid_t ntype, hsize_t count[], void *databuf);

hid_t h4h5_PTopen(const char *filename, uintn access);
herr_t h4h5_PTclose(hid_t fid);
long h4h5_PTinqpoint(const char* filename, char *pointlist, long *strbufsize);

long h4h5_EHinqglbattrs(hid_t fileid, char *attnames, long *strbufsize);
herr_t h4h5_EHreadglbattr(hid_t fileid, const char *attname, void *databuf);
herr_t h4h5_EHwriteglbattr(hid_t fileid, const char *attname, hid_t ntype, hsize_t count[], void *databuf);
herr_t h4h5_EHwriteodl(hid_t hdfeos_handle, const char *groupname, char *odlstring);

/* Generic type mapping and conversion stuff */
/* HDF-EOS type inquiry calls can return full-blown HDF5 type numbers
   (endian-info and everything).  These functions map those types onto
   the types specific to your platform, and disassemble those types into
   int/float/string size signed info
*/
void int32_to_hsize_t(hsize_t *f, long fsize);
int32 *new_int32(const hsize_t *f, long fsize);
hid_t read_in_type(hid_t type);
H5T_class_t type_info (hid_t type, int *size, int *signp);
hid_t h4type_to_h5type(hid_t h4type);
hid_t h5type_to_h4type(hid_t h5type);
hid_t hdfeos_native_to_hdf_native(hid_t heval);
int count_char(char c, char const *s);

#endif
