/* The actual code for merging HDF-EOS 4/5 API calls.  I'm going to
   try to make everything look like HDF-EOS 5.  It'll all *look* like
   HDF-EOS 5, but you won't be able to rely on HDF-EOS 5 only functionality.

   This thing is full of evil casts to shut the compiler up.  Many of them
   are "harmless" (HE5 used const, HE4 didn't, but promises not to modify).
   All potentially dangerous casts are marked with /**/
/* Note the brilliant design decision of making C's comments not nest. */

#include "he25.h"

int h4orh5 = 0;

int he25which() { return h4orh5; }

int he25choose(int which)
{
  switch(which) {
  case H4FILE:
  case H5FILE:
    h4orh5 = which;
    break;
  default:
    which = 0;
    break;
  }
  return which;
}
    

/* Swath compatibility layer */

hid_t he25_SWopen(const char *filename, uintn access)
{
  hid_t result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_SWopen(filename, access);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }
  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    intn he4access;

    if (access ==  H5F_ACC_RDWR)
      he4access = DFACC_RDWR;
    else if (access ==  H5F_ACC_RDONLY)
      he4access = DFACC_READ;
    else if (access ==  H5F_ACC_TRUNC)
      he4access = DFACC_CREATE;
    else
      return FAIL;

    if((result = SWopen((char *)filename, he4access)) != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}
hid_t he25_SWcreate(hid_t swhandle, const char *swath_name)
{
  switch(he25which()) {
  case H4FILE:
    return SWcreate(swhandle, (char *)swath_name);
  case H5FILE:
    return HE5_SWcreate(swhandle, swath_name);
  default:
    return FAIL;
  }
}

long he25_SWinqswath(const char* filename, char *swathlist, long *strbufsize)
{
  long result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_SWinqswath(filename, swathlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }
  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    result = SWinqswath((char *)filename, swathlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}

hid_t he25_SWattach(hid_t fid, const char *swathname)
{
  switch(he25which()) {
  case H4FILE:
    return SWattach(fid, (char *)swathname);
  case H5FILE:
    return HE5_SWattach(fid, swathname);
  default:
    return FAIL;
  }
}

herr_t he25_SWclose(hid_t fid)
{
  switch(he25which()) {
  case H4FILE:
    h4orh5 = 0;
    return SWclose(fid);
  case H5FILE:
    h4orh5 = 0;
    return HE5_SWclose(fid);
  default:
    return FAIL;
  }
}

herr_t he25_SWdetach(hid_t swhandle)
{
  switch(he25which()) {
  case H4FILE:
    return SWdetach(swhandle);
  case H5FILE:
    return HE5_SWdetach(swhandle);
  default:
    return FAIL;
  }
}

long he25_SWinqattrs(hid_t swhandle, char *attrlist, long *strbufsize)
{
  switch(he25which()) {
  case H4FILE:
    return SWinqattrs(swhandle, attrlist, strbufsize);
  case H5FILE:
    return HE5_SWinqattrs(swhandle, attrlist, strbufsize);
  default:
    return FAIL;
  }
}

long he25_SWinqdims(hid_t swhandle, char *dimnamebuf, hsize_t *dimsizebuf)
{
  switch(he25which()) {
  case H4FILE:
    {
      long result = SWinqdims(swhandle, dimnamebuf, /**/(int32 *)dimsizebuf);
      if (result >= 0)
	int32_to_hsize_t(dimsizebuf, result);
      return result;
    }
  case H5FILE:
    return HE5_SWinqdims(swhandle, dimnamebuf, dimsizebuf);
  default:
    return FAIL;
  }
}

long he25_SWinqmaps(hid_t swhandle, char *dimmap, long *offset, long *increment)
{
  switch(he25which()) {
  case H4FILE:
    return SWinqmaps(swhandle, dimmap, offset, increment);
  case H5FILE:
    return HE5_SWinqmaps(swhandle, dimmap, offset, increment);
  default:
    return FAIL;
  }
}

/* As it turns out, the entrycodes for H4 and H5 match up */

long he25_SWnentries(hid_t swhandle, int entrycode, long *strbufsize)
{
  switch(he25which()) {
  case H4FILE:
    return SWnentries(swhandle, entrycode, strbufsize);
  case H5FILE:
    return HE5_SWnentries(swhandle, entrycode, strbufsize);
  default:
    return FAIL;
  }
}

herr_t he25_SWreadattr(hid_t swhandle, char *attrname, void *datbuf)
{
  switch(he25which()) {
  case H4FILE:
    return SWreadattr(swhandle, attrname, datbuf);
  case H5FILE:
    return HE5_SWreadattr(swhandle, attrname, datbuf);
  default:
    return FAIL;
  }
}

herr_t he25_SWreadfield(hid_t swhandle, char *fieldname, const hssize_t *start, const hsize_t *stride, const hsize_t *edge, void *datbuf)
{
  switch(he25which()) {
  case H4FILE:
    /* If any of start, stride, or edge are non-NULL here, they must
       be converted from hsize_t to int32 */
    if(start!= NULL || stride != NULL || edge != NULL) {
      long numdim;
      long i;
      int32 *cstart = NULL;
      int32 *cstride = NULL;
      int32 *cedge = NULL;
      herr_t result;
      int32 bufsize;

      numdim = SWnentries(swhandle, HDFE_NENTDIM, &bufsize);
      if(start != NULL)
	cstart = new_int32(start, numdim);
      if(stride != NULL)
	cstride = new_int32(stride, numdim);
      if(edge != NULL)
	cedge = new_int32(edge, numdim);

      result = SWreadfield(swhandle, fieldname, cstart, cstride, cedge, datbuf);
      if(cstart != NULL) free(cstart);
      if(cstride != NULL) free(cstride);
      if(cedge!= NULL) free(cedge);
      return result;
    }
    else
      return SWreadfield(swhandle, fieldname, NULL, NULL, NULL, datbuf);

  case H5FILE:
    return HE5_SWreadfield(swhandle, fieldname, start, stride, edge, datbuf);
  default:
    return FAIL;
  }
}


herr_t he25_SWattrinfo(hid_t swhandle, const char *attrname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* It is necessary to remap the ntype value here */
    {
      herr_t result = SWattrinfo(swhandle, /**/(char *)attrname, /**/(int32 *)ntype, /**/(int32 *)count);
      if(result >= 0) {
	if(ntype != NULL) {
	  *ntype = h4type_to_h5type(*ntype);
	  int32_to_hsize_t(count, 1);
	}
      }
      return result;
    }
  case H5FILE:
    return HE5_SWattrinfo(swhandle, attrname, ntype, count);
  default:
    return FAIL;
  }
}

/* H4 has no equivalent for local/group attr info, so we punt and say there
   isn't any */

herr_t he25_SWlocattrinfo(hid_t swhandle, const char *fieldname, const char *attrname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* We can't do this in H4, so return that there aren't any */
    *count = 0;
    return SUCCEED;

  case H5FILE:
    return HE5_SWlocattrinfo(swhandle, fieldname, attrname, ntype, count);
  default:
    return FAIL;
  }
}

herr_t he25_SWgrpattrinfo(hid_t swhandle, const char *attrname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* We can't do this in H4, so return that there aren't any */
    *count = 0;
    return SUCCEED;

  case H5FILE:
    return HE5_SWgrpattrinfo(swhandle, attrname, ntype, count);
  default:
    return FAIL;
  }
}

herr_t he25_SWfieldinfo(hid_t swhandle, char *fieldname, int *rank, hsize_t *dims, H5T_class_t *ntype, char *dimlist, char *maxdimlist)
{
  switch(he25which()) {
  case H4FILE:
    {
      /* It may be necessary to remap the ntype value here */
      /* and we just ignore/punt on the maxdimlist for HE4 */

      herr_t result = SWfieldinfo(swhandle, fieldname, /**/(int32 *)rank, /**/(int32 *)dims, /**/(int32 *)ntype, dimlist);

      if(result >= 0) {
	if(ntype != NULL) {
	  *ntype = h4type_to_h5type(*ntype);

	  int32_to_hsize_t(dims, *rank);
	}
      }
      return result;
    }
  case H5FILE:
    return HE5_SWfieldinfo(swhandle, fieldname, rank, dims, ntype, dimlist, maxdimlist);
  default:
    return FAIL;
  }
}

herr_t he25_SWinqdatafields(hid_t swhandle, char *fieldlist, int *rank, H5T_class_t *ntype)
{
  switch(he25which()) {
  case H4FILE:
    /* It may be necessary to remap the ntype value here */
    {
      herr_t result = SWinqdatafields(swhandle, fieldlist, /**/(int32 *)rank, /**/(int32 *)ntype);
      if(result >= 0) {
	if(ntype != NULL) {
	  long fieldcount = count_char(',', fieldlist) + 1;
	  while(fieldcount > 0) {
	    hid_t H5ntype = h4type_to_h5type(ntype[--fieldcount]);
	    ntype[fieldcount] = H5ntype;
	  }
	}
	/* int32_to_hsize_t(ntype, fieldcount); */
      }
      return result;
    }
  case H5FILE:
    return HE5_SWinqdatafields(swhandle, fieldlist, rank, ntype);
  default:
    return FAIL;
  }
}

herr_t he25_SWinqgeofields(hid_t swhandle, char *fieldlist, int *rank, H5T_class_t *ntype)
{
  switch(he25which()) {
  case H4FILE:
    /* It may be necessary to remap the ntype value here */
    {
      herr_t result = SWinqgeofields(swhandle, fieldlist, /**/(int32 *)rank, /**/(int32 *)ntype);
      if(result >= 0) {
	if(ntype != NULL) {
	  long fieldcount = count_char(',', fieldlist) + 1;
	  while(fieldcount > 0) {
	    hid_t H5ntype = h4type_to_h5type(ntype[--fieldcount]);
	    ntype[fieldcount] = H5ntype;
	  }
	}
	/* int32_to_hsize_t(ntype, fieldcount); */
      }
      return result;
    }
  case H5FILE:
    return HE5_SWinqgeofields(swhandle, fieldlist, rank, ntype);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefcomp(hid_t swhandle, int compcode, int *compparm)
{
  /* Codes and parameters in common match */
  switch(he25which()) {
  case H4FILE:
    return SWdefcomp(swhandle, compcode, compparm);
  case H5FILE:
    return HE5_SWdefcomp(swhandle, compcode, compparm);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefdatafield(hid_t swhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge)
{
  /* Need to map types? */
  switch(he25which()) {
  case H4FILE:
    return SWdefdatafield(swhandle, (char *)fieldname, dimlist, h5type_to_h4type(ntype), merge);
  case H5FILE:
    return HE5_SWdefdatafield(swhandle, (char *)fieldname, dimlist, maxdimlist, hdfeos_native_to_hdf_native(ntype), merge);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefdim(hid_t swhandle, char *dimname, hsize_t dimsize)
{
  switch(he25which()) {
  case H4FILE:
    return SWdefdim(swhandle, dimname, dimsize);
  case H5FILE:
    return HE5_SWdefdim(swhandle, dimname, dimsize);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefdimmap(hid_t swhandle, char *geodim, char *datadim, hsize_t offset, hsize_t increment)
{
  switch(he25which()) {
  case H4FILE:
    return SWdefdimmap(swhandle, geodim, datadim, offset, increment);
  case H5FILE:
    return HE5_SWdefdimmap(swhandle, geodim, datadim, offset, increment);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefgeofield(hid_t swhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge)
{
  /* Need to map types? */
  switch(he25which()) {
  case H4FILE:
    return SWdefgeofield(swhandle, (char *)fieldname, dimlist,  h5type_to_h4type(ntype), merge);
  case H5FILE:
    return HE5_SWdefgeofield(swhandle, (char *)fieldname, dimlist, maxdimlist, hdfeos_native_to_hdf_native(ntype), merge);
  default:
    return FAIL;
  }
}

herr_t he25_SWdefidxmap(hid_t swhandle, char *geodim, char *datadim, long index[])
{
  switch(he25which()) {
  case H4FILE:
    return SWdefidxmap(swhandle, geodim, datadim, index);
  case H5FILE:
    return HE5_SWdefidxmap(swhandle, geodim, datadim, index);
  default:
    return FAIL;
  }
}

herr_t he25_SWwriteattr(hid_t swhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  /* Do we need to map the type from H5 to H4 here? */
  /* We have to dereference the count array */
  switch(he25which()) {
  case H4FILE:
    return SWwriteattr(swhandle, (char *)attrname, h5type_to_h4type(ntype), count[0], databuf);
  case H5FILE:
    return HE5_SWwriteattr(swhandle, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_SWwritefield(hid_t swhandle, char *fieldname,  const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    /* If any of start, stride, or edge are non-NULL here, they must
       be converted from hsize_t to int32 */
    if(start!= NULL || stride != NULL || edge != NULL){
      long numdim;
      long i;
      int32 *cstart = NULL;
      int32 *cstride = NULL;
      int32 *cedge = NULL;
      herr_t result;
      int32 bufsize;

      numdim = SWnentries(swhandle, HDFE_NENTDIM, &bufsize);
      if(start != NULL)
	cstart = new_int32(start, numdim);
      if(stride != NULL)
	cstride = new_int32(stride, numdim);
      if(edge != NULL)
	cedge = new_int32(edge, numdim);

      result = SWwritefield(swhandle, (char *)fieldname, cstart, cstride, cedge, databuf);
      if(cstart != NULL) free(cstart);
      if(cstride != NULL) free(cstride);
      if(cedge!= NULL) free(cedge);
      return result;
    }
    else
      return SWwritefield(swhandle, (char *)fieldname, NULL, NULL, NULL, databuf);

  case H5FILE:
    return HE5_SWwritefield(swhandle, fieldname, start, stride, edge, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_SWwritelocattr(hid_t swhandle, const char *fieldname, char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_SWwritelocattr(swhandle, fieldname, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_SWwritegrpattr(hid_t swhandle, char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_SWwritegrpattr(swhandle, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_SWsetalias(hid_t swhandle, char *fieldname, const char *aliaslist)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_SWsetalias(swhandle, fieldname, aliaslist);
  default:
    return FAIL;
  }
}

/* And the Profiling crap */

herr_t he25_PRdefine(hid_t swhandle, const char *prname, char *dimlist, char *maxdimlist, hid_t datatype)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_PRdefine(swhandle, prname, dimlist, maxdimlist, hdfeos_native_to_hdf_native(datatype));
  default:
    return FAIL;
  }
}

herr_t he25_PRinfo(hid_t swhandle, const char *prname, int *rank, hsize_t dims[], hsize_t maxdims[], H5T_class_t *classID, char *dimlist, char *maxdimlist)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_PRinfo(swhandle, prname, rank, dims, maxdims, classID, dimlist, maxdimlist);
  default:
    return FAIL;
  }
}

herr_t he25_PRinquire(hid_t swhandle, char *prnames, int *rank, H5T_class_t *classID)
{
  switch(he25which()) {
  case H4FILE:
    /* H4 files have no profiles */
    if(prnames != NULL) *prnames = 0;
    return 0;
  case H5FILE:
    return HE5_PRinquire(swhandle, prnames, rank, classID);
  default:
    return FAIL;
  }
}

herr_t he25_PRread(hid_t swhandle, const char *prname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_PRread(swhandle, prname, start, stride, edge, databuf);
  default:
    return FAIL;
  }
}

/* I'm wondering why this needs to exist, and if it does, why isn't it
   also necessary for swaths?
*/
herr_t he25_PRreclaimspace(hid_t swhandle, const char *prname, void *buffer)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_PRreclaimspace(swhandle, prname, buffer);
  default:
    return FAIL;
  }
}

herr_t he25_PRwrite(hid_t swhandle, const char *prname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], size_t size, void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_PRwrite(swhandle, prname, start, stride, edge, size, databuf);
  default:
    return FAIL;
  }
}


/* Grid compatibility layer */

hid_t he25_GDopen(const char *filename, uintn access)
{
  hid_t result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_GDopen(filename, access);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }
  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    intn he4access;

    h4orh5 = 0;
    if (access ==  H5F_ACC_RDWR)
      he4access = DFACC_RDWR;
    else if (access ==  H5F_ACC_RDONLY)
      he4access = DFACC_READ;
    else if (access ==  H5F_ACC_TRUNC)
      he4access = DFACC_CREATE;
    else
      return FAIL;

    if((result = GDopen((char *)filename, he4access)) != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}

hid_t he25_GDcreate(hid_t fid, const char *gridname, long xdimsize, long ydimsize, double upleft[], double upright[])
{
  switch(he25which()) {
  case H4FILE:
    return GDcreate(fid, (char *)gridname, xdimsize, ydimsize, upleft, upright);
  case H5FILE:
    return HE5_GDcreate(fid, gridname, xdimsize, ydimsize, upleft, upright);
  default:
    return FAIL;
  }
}
long he25_GDinqgrid(const char* filename, char *gridlist, long *strbufsize)
{
  long result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_GDinqgrid(filename, gridlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }

  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    result = GDinqgrid((char *)filename, gridlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}

hid_t he25_GDattach(hid_t fid, const char *gridname)
{
  switch(he25which()) {
  case H4FILE:
    return GDattach(fid, (char *)gridname);
  case H5FILE:
    return HE5_GDattach(fid, gridname);
  default:
    return FAIL;
  }
}

herr_t he25_GDattrinfo(hid_t grhandle, const char *attname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* It may be necessary to remap the ntype value here */
    {
      herr_t result = GDattrinfo(grhandle, (char *)attname, /**/(int32 *)ntype, /**/(int32 *)count);
      if(result >= 0) {
	if(ntype != NULL) {
	  *ntype = h4type_to_h5type(*ntype);
	  int32_to_hsize_t(count, 1);
	}
      }
      return result;
    }
  case H5FILE:
    return HE5_GDattrinfo(grhandle, attname, ntype, count);
  default:
    return FAIL;
  }
}

/* There is no H4 equivalent inquiry for this, so we punt and say there
   aren't any */

herr_t he25_GDlocattrinfo(hid_t grhandle, const char *fieldname, const char *attname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* We punt and say there aren't any */
    *count = 0;
    return SUCCEED;
  case H5FILE:
    return HE5_GDlocattrinfo(grhandle, fieldname, attname, ntype, count);
  default:
    return FAIL;
  }
}
herr_t he25_GDgrpattrinfo(hid_t grhandle, const char *attname, H5T_class_t *ntype, hsize_t *count)
{
  switch(he25which()) {
  case H4FILE:
    /* We punt and say there aren't any */
    *count = 0;
    return SUCCEED;
  case H5FILE:
    return HE5_GDgrpattrinfo(grhandle, attname, ntype, count);
  default:
    return FAIL;
  }
}

herr_t he25_GDclose(hid_t fid)
{
  switch(he25which()) {
  case H4FILE:
    h4orh5 = 0;
    return GDclose(fid);
  case H5FILE:
    h4orh5 = 0;
    return HE5_GDclose(fid);
  default:
    return FAIL;
  }
}

herr_t he25_GDdetach(hid_t grhandle)
{
  switch(he25which()) {
  case H4FILE:
    return GDdetach(grhandle);
  case H5FILE:
    return HE5_GDdetach(grhandle);
  default:
    return FAIL;
  }
}

herr_t he25_GDfieldinfo(hid_t grhandle, char *fieldname, int *rank, hsize_t *dims, H5T_class_t *ntype, char *dimlist, char *maxdimlist)
{
  switch(he25which()) {
  case H4FILE:
    {
      /* It may be necessary to remap the ntype value here */
      /* and we just ignore/punt on the maxdimlist for HE4 */

      herr_t result = GDfieldinfo(grhandle, fieldname, /**/(int32 *)rank, /**/(int32 *)dims, /**/(int32 *)ntype, dimlist);

      if(result >= 0) {
	if(ntype != NULL) {
	  *ntype = h4type_to_h5type(*ntype);

	  int32_to_hsize_t(dims, *rank);
	}
      }
      return result;
    }
  case H5FILE:
    return HE5_GDfieldinfo(grhandle, fieldname, rank, dims, ntype, dimlist, maxdimlist);
  default:
    return FAIL;
  }
}

herr_t he25_GDgridinfo(hid_t grhandle, long *xsize, long *ysize, double *upleft, double *lowright)
{
  switch(he25which()) {
  case H4FILE:
    return GDgridinfo(grhandle, xsize, ysize, upleft, lowright);
  case H5FILE:
    return HE5_GDgridinfo(grhandle, xsize, ysize, upleft, lowright);
  default:
    return FAIL;
  }
}

long he25_GDinqattrs(hid_t grhandle, char *attrlist, long *strbufsize)
{
  switch(he25which()) {
  case H4FILE:
    return GDinqattrs(grhandle, attrlist, strbufsize);
  case H5FILE:
    return HE5_GDinqattrs(grhandle, attrlist, strbufsize);
  default:
    return FAIL;
  }
}

herr_t he25_GDinqfields(hid_t grhandle, char *fieldlist, int *rank, H5T_class_t *ntype)
{
  switch(he25which()) {
  case H4FILE:
    /* It may be necessary to remap the ntype value here */
    {
      herr_t result = GDinqfields(grhandle, fieldlist, /**/(int32 *)rank, /**/(int32 *)ntype);
      if(result >= 0) {
	if(ntype != NULL) {
	  long fieldcount = count_char(',', fieldlist) + 1;
	  while(fieldcount > 0) {
	    hid_t H5ntype = h4type_to_h5type(ntype[--fieldcount]);
	    ntype[fieldcount] = H5ntype;
	  }
	}
	/* int32_to_hsize_t(ntype, fieldcount); */
      }
      return result;
    }
  case H5FILE:
    return HE5_GDinqfields(grhandle, fieldlist, rank, ntype);
  default:
    return FAIL;
  }
}

long he25_GDinqdims(hid_t grhandle, char *dimnamebuf, hsize_t *dimsizebuf)
{
  switch(he25which()) {
  case H4FILE:
    {
      long result = GDinqdims(grhandle, dimnamebuf, /**/(int32 *)dimsizebuf);
      if (result >= 0)
	int32_to_hsize_t(dimsizebuf, result);
      return result;
    }
  case H5FILE:
    return HE5_GDinqdims(grhandle, dimnamebuf, dimsizebuf);
  default:
    return FAIL;
  }
}

/* As it turns out, the entrycodes for H4 and H5 match up */

long he25_GDnentries(hid_t grhandle, int entrycode, long *strbufsize)
{
  switch(he25which()) {
  case H4FILE:
    return GDnentries(grhandle, entrycode, strbufsize);
  case H5FILE:
    return HE5_GDnentries(grhandle, entrycode, strbufsize);
  default:
    return FAIL;
  }
}

herr_t he25_GDprojinfo(hid_t grhandle, int *projcode, int *zonecode, int *spherecode, double *projparm)
{
  switch(he25which()) {
  case H4FILE:
    {
      int32 pc, zc, sc;
      herr_t result = GDprojinfo(grhandle, &pc, &zc, &sc, projparm);
      *projcode = pc;
      *zonecode = zc;
      *spherecode = sc;
      return result;
    }

  case H5FILE:
    return HE5_GDprojinfo(grhandle, projcode, zonecode, spherecode, projparm);
  default:
    return FAIL;
  }
}

herr_t he25_GDreadattr(hid_t grhandle, char *attrname, void *datbuf)
{
  switch(he25which()) {
  case H4FILE:
    return GDreadattr(grhandle, attrname, datbuf);
  case H5FILE:
    return HE5_GDreadattr(grhandle, attrname, datbuf);
  default:
    return FAIL;
  }
}


herr_t he25_GDreadfield(hid_t grhandle, char *fieldname, const hssize_t *start, const hsize_t *stride, const hsize_t *edge, void *datbuf)
{
  switch(he25which()) {
  case H4FILE:
    /* If any of start, stride, or edge are non-NULL here, they must
       be converted from hsize_t to int32 */
    if(start!= NULL || stride != NULL || edge != NULL){
      long numdim;
      long i;
      int32 *cstart = NULL;
      int32 *cstride = NULL;
      int32 *cedge = NULL;
      herr_t result;
      int32 bufsize;

      numdim = GDnentries(grhandle, HDFE_NENTDIM, &bufsize);
      if(start != NULL)
	cstart = new_int32(start, numdim);
      if(stride != NULL)
	cstride = new_int32(stride, numdim);
      if(edge != NULL)
	cedge = new_int32(edge, numdim);

      result = GDreadfield(grhandle, fieldname, cstart, cstride, cedge, datbuf);
      if(cstart != NULL) free(cstart);
      if(cstride != NULL) free(cstride);
      if(cedge!= NULL) free(cedge);
      return result;
    }
    else
      return GDreadfield(grhandle, fieldname, NULL, NULL, NULL, datbuf);

  case H5FILE:
    return HE5_GDreadfield(grhandle, fieldname, start, stride, edge, datbuf);
  default:
    return FAIL;
  }
}

herr_t he25_GDdefcomp(hid_t grhandle, int dimname, int dimsize[])
{
  switch(he25which()) {
  case H4FILE:
    return GDdefcomp(grhandle, dimname, dimsize);
  case H5FILE:
    return HE5_GDdefcomp(grhandle, dimname, dimsize);
  default:
    return FAIL;
  }
}

/* There is no equivalent HDF-EOS4 call for this function, dangit! */
herr_t he25_GDdefcomtile(hid_t grhandle, int compcode, int *compparm, int tilerank, const hsize_t *tiledim)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_GDdefcomtile(grhandle, compcode, compparm, tilerank, tiledim);
  default:
    return FAIL;
  }
}

herr_t he25_GDdefdim(hid_t grhandle, char *dimname, hsize_t dim)
{
  switch(he25which()) {
  case H4FILE:
    return GDdefdim(grhandle, dimname, dim);
  case H5FILE:
    return HE5_GDdefdim(grhandle, dimname, dim);
  default:
    return FAIL;
  }
}

herr_t he25_GDdeffield(hid_t grhandle, const char *fieldname, char *dimlist, char *maxdimlist, hid_t ntype, int merge)
{
  switch(he25which()) {
  case H4FILE:
    /* Do we need to map the type from H5 to H4 here? */
    /* We do need to blithely ignore the maxdimlist... or maybe
       complain if it's supplied? */
    return GDdeffield(grhandle, (char *)fieldname, dimlist, h5type_to_h4type(ntype), merge);
  case H5FILE:
    return HE5_GDdeffield(grhandle, fieldname, dimlist, maxdimlist, hdfeos_native_to_hdf_native(ntype), merge);
  default:
    return FAIL;
  }
}

herr_t he25_GDdeforigin(hid_t grhandle, int origincode)
{
  /* The origincodes match up */
  switch(he25which()) {
  case H4FILE:
    return GDdeforigin(grhandle, origincode);
  case H5FILE:
    return HE5_GDdeforigin(grhandle, origincode);
  default:
    return FAIL;
  }
}

herr_t he25_GDdefpixreg(hid_t grhandle, int prcode)
{
  /* The prcodes match up */
  switch(he25which()) {
  case H4FILE:
    return GDdefpixreg(grhandle, prcode);
  case H5FILE:
    return HE5_GDdefpixreg(grhandle, prcode);
  default:
    return FAIL;
  }
}

herr_t he25_GDdefproj(hid_t grhandle, int projcode, int zonecode, int spherecode, double projparm[])
{
  /* The codes match up */
  switch(he25which()) {
  case H4FILE:
    return GDdefproj(grhandle, projcode, zonecode, spherecode, projparm);
  case H5FILE:
    return HE5_GDdefproj(grhandle, projcode, zonecode, spherecode, projparm);
  default:
    return FAIL;
  }
}

herr_t he25_GDdeftile(hid_t grhandle, int tilecode, int tilerank, const hsize_t tiledims[])
{
  /* The codes match up, but we may have to mung the dimension array */
  switch(he25which()) {
  case H4FILE: {
    int32 *ctiledims = NULL;
    herr_t result;
    if(tiledims != NULL)
      ctiledims = new_int32(tiledims, tilerank);
    result = GDdeftile(grhandle, tilecode, tilerank, ctiledims);
    if(ctiledims != NULL) free(ctiledims);
    return result;
  }
  case H5FILE:
    return HE5_GDdeftile(grhandle, tilecode, tilerank, tiledims);
  default:
    return FAIL;
  }
}

herr_t he25_GDwriteattr(hid_t grhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  /* Do we need to map the type from H5 to H4 here? */
  /* We have to dereference the count array */
  switch(he25which()) {
  case H4FILE:
    return GDwriteattr(grhandle, (char *)attrname, h5type_to_h4type(ntype), count[0], databuf);
  case H5FILE:
    return HE5_GDwriteattr(grhandle, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_GDwritefield(hid_t grhandle, const char *fieldname, const hssize_t start[], const hsize_t stride[], const hsize_t edge[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    /* If any of start, stride, or edge are non-NULL here, they must
       be converted from hsize_t to int32 */
    if(start!= NULL || stride != NULL || edge != NULL){
      long numdim;
      long i;
      int32 *cstart = NULL;
      int32 *cstride = NULL;
      int32 *cedge = NULL;
      herr_t result;
      int32 bufsize;

      numdim = GDnentries(grhandle, HDFE_NENTDIM, &bufsize);
      if(start != NULL)
	cstart = new_int32(start, numdim);
      if(stride != NULL)
	cstride = new_int32(stride, numdim);
      if(edge != NULL)
	cedge = new_int32(edge, numdim);

      result = GDwritefield(grhandle, (char *)fieldname, cstart, cstride, cedge, databuf);
      if(cstart != NULL) free(cstart);
      if(cstride != NULL) free(cstride);
      if(cedge!= NULL) free(cedge);
      return result;
    }
    else
      return GDwritefield(grhandle, (char *)fieldname, NULL, NULL, NULL, databuf);

  case H5FILE:
    return HE5_GDwritefield(grhandle, fieldname, start, stride, edge, databuf);
  default:
    return FAIL;
  }
}

/* H4 doesn't have group/local attributes, so these writes just fail */

herr_t he25_GDwritegrpattr(hid_t grhandle, const char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_GDwritegrpattr(grhandle, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_GDwritelocattr(hid_t grhandle, const char *fieldname, const char *attrname, hid_t ntype, hsize_t count[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_GDwritelocattr(grhandle, fieldname, attrname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

/* Point compatibility layer (minimal - very little need to support Points,
   judging from the number of files containing them */

hid_t he25_PTopen(const char *filename, uintn access)
{
  hid_t result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_PTopen(filename, access);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }
  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    intn he4access;

    if (access ==  H5F_ACC_RDWR)
      he4access = DFACC_RDWR;
    else if (access ==  H5F_ACC_RDONLY)
      he4access = DFACC_READ;
    else if (access ==  H5F_ACC_TRUNC)
      he4access = DFACC_CREATE;
    else
      return FAIL;

    if((result = PTopen((char *)filename, he4access)) != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}

herr_t he25_PTclose(hid_t fid)
{
  switch(he25which()) {
  case H4FILE:
    h4orh5 = 0;
    return PTclose(fid);
  case H5FILE:
    h4orh5 = 0;
    return HE5_PTclose(fid);
  default:
    return FAIL;
  }
}

long he25_PTinqpoint(const char* filename, char *pointlist, long *strbufsize)
{
  long result = 0;

  if(h4orh5 == 0 || h4orh5 == H5FILE) {
    result = HE5_PTinqpoint(filename, pointlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H5FILE;
  }
  if(h4orh5 == 0 || h4orh5 == H4FILE) {
    result = PTinqpoint((char *)filename, pointlist, strbufsize);
    if(result != FAIL)
      h4orh5 = H4FILE;
  }
  return result;
}

/* Global attribute stuff
   H4 files don't have global attributes, so:
   * inquiries succeed, but say "no attributes"
   * read/write attempts fail
   */
long he25_EHinqglbattrs(hid_t fileid, char *attnames, long *strbufsize)
{
 switch(he25which()) {
  case H4FILE:
    if(attnames != NULL)
      *attnames = 0;
    if(strbufsize != NULL)
      *strbufsize = 0;
    return 0;
    break;
  case H5FILE:
    return HE5_EHinqglbattrs(fileid, attnames, strbufsize);
    break;
  default:
    return FAIL;
  }
}

herr_t he25_EHreadglbattr(hid_t fileid, const char *attname, void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return he25_EHreadglbattr(fileid, attname, databuf);
  default:
    return FAIL;
  }
}

herr_t he25_EHwriteglbattr(hid_t fileid, const char *attname, hid_t ntype, hsize_t count[], void *databuf)
{
  switch(he25which()) {
  case H4FILE:
    return FAIL;
  case H5FILE:
    return HE5_EHwriteglbattr(fileid, (char *)attname, hdfeos_native_to_hdf_native(ntype), count, databuf);
  default:
    return FAIL;
  }
}

/* Additional stuff with no analog in HDF-EOS 4/5 API (but there should be)
   to write the ODL metadata out.  Should also do a reader, no doubt... */

/* Snarfed whole from Zhangshi's hemu to write metadata */

/* Write dataset. */
int WriteDataset5(char *MetadataNamen, char *MetaBuffer, size_t TotalSize, hid_t gid);
int WriteDataset5(char *MetadataNamen, char *MetaBuffer, size_t TotalSize, hid_t gid)
{
  hid_t did, dtype, dspace;
  hsize_t dims[1];

  /*  Create new metadata. */
  dtype = H5Tcopy (H5T_C_S1);
  if ((H5Tset_size (dtype, TotalSize))<0)
  {
    fprintf(stderr, "Cannot set size.\n");
    exit(1);
  }
  dims[0] = 1;
  dspace = H5Screate_simple(1, dims, NULL);
  if ((did = H5Dcreate(gid, MetadataNamen, dtype, dspace, H5P_DEFAULT))<0)
  {
    fprintf(stderr, "Cannot create metadata.\n");
    exit(1);
  }

  /*  Write new metadata. */
  TRACE((stderr,"Writing metadata: %s.\n", MetadataNamen));
  if (H5Dwrite(did, dtype, dspace, H5S_ALL, H5P_DEFAULT, MetaBuffer) < 0 ) 
  {
    fprintf(stderr,"Cannot write to dataset %s.\n", MetadataNamen);
    exit(1);
  }

  /*  Close dataspace. */
  if (H5Sclose(dspace) == -1)
  {
    fprintf(stderr,"Cannot close dataspace for %s.\n", MetadataNamen);
    exit(1);
  }

  /*  Close datatype. */
  if (H5Tclose(dtype) == -1)
  {
    fprintf(stderr,"Cannot close datatype for %s.\n", MetadataNamen);
    exit(1);
  }

  /*  Close Dataset.  */
  if (H5Dclose(did) == -1)
  {
    fprintf(stderr,"Cannot close dataset for %s.\n", MetadataNamen);
    exit(1);
  }

  return (0);
}

herr_t he25_EHwriteodl(hid_t hdfeos_handle, const char *groupname, char *odlstring)
{
  switch(he25which()) {
  case H4FILE:
    {
      //      hid_t hdf_handle;
      hid_t gid;
      long remaining = strlen(odlstring);
      int chunk_number = 0;
      const int CHUNK_SIZE = 65535;

      /* HORRIBLE HACK to follow - since we can't map the hdfeos_handle
	 to its corresponding hdf_handle with the API, we're going to
	 * get the filename out with EHfilename
	 * close the hdfeos_handle
	 * open an hdf_handle
	 * mung the file
	 * close the hdf_handle
	 * reopen the hdfeos_handle (and spit if we don't get back the one we closed)

	 YECCH.
      */
      char hdfeos_file_name[1024];
      if(EHfilename(hdfeos_handle, hdfeos_file_name) < 0)
	return FAIL;
      he25_SWclose(hdfeos_handle);
      //      if((hdf_handle = H5Fopen(hdfeos_file_name, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
      //	return FAIL;

      if((gid = SDstart(hdfeos_file_name, DFACC_WRITE)) < 0)
	return FAIL;

      /* Do we need to delete any existing metadata here?  For the moment,
	 I'm assuming NO */

      /* We get to loop down the odlstring writing out the metadata in 64K
	 chunks.  The most expedient way to do that is to clobber one character
	 to null out the substring in-place.  Yecch. */
      while(remaining > 0) {
	char null_memory;
	char chunk_name[100];

	if(remaining > CHUNK_SIZE) {
	  null_memory = odlstring[CHUNK_SIZE];
	  odlstring[CHUNK_SIZE] = 0;
	}

	sprintf(chunk_name, "%s.%d", groupname, chunk_number);
	if(SDsetattr(gid, chunk_name, DFNT_CHAR, (remaining > CHUNK_SIZE ? CHUNK_SIZE : remaining), odlstring) < 0)
	  return FAIL;

	if(remaining > CHUNK_SIZE) {
	  odlstring[CHUNK_SIZE] = null_memory;
	}
	remaining -= CHUNK_SIZE;
	odlstring += CHUNK_SIZE;
      }
    
      /* We're done, close the group */
      if(SDend(gid) < 0)
	return FAIL;

      /* Close and reopen hdfeos_handle */
      //      H5Fclose(hdf_handle);
      if(hdfeos_handle != he25_SWopen(hdfeos_file_name, H5F_ACC_RDWR))
	return FAIL;
    }
    TRACE((stderr, "ODL for HDF4 %s:\n%s\n", groupname, odlstring));
    return 0;
  case H5FILE:
    {
      hid_t hdf_handle;
      hid_t gid;
      long remaining = strlen(odlstring);
      int chunk_number = 0;
      const int CHUNK_SIZE = 65535;

      /* HORRIBLE HACK to follow - since we can't map the hdfeos_handle
	 to its corresponding hdf_handle with the API, we're going to
	 * get the filename out with HE5_EHfilename
	 * close the hdfeos_handle
	 * open an hdf_handle
	 * mung the file
	 * close the hdf_handle
	 * reopen the hdfeos_handle (and spit if we don't get back the one we closed)

	 YECCH.
      */
      char hdfeos_file_name[1024];
      if(HE5_EHfilename(hdfeos_handle, hdfeos_file_name) < 0)
	return FAIL;
      he25_SWclose(hdfeos_handle);
      if((hdf_handle = H5Fopen(hdfeos_file_name, H5F_ACC_RDWR, H5P_DEFAULT)) < 0)
	return FAIL;

      if((gid = H5Gopen(hdf_handle, "HDFEOS INFORMATION")) < 0)
	return FAIL;

      /* Do we need to delete any existing metadata here?  For the moment,
	 I'm assuming NO */

      /* We get to loop down the odlstring writing out the metadata in 64K
	 chunks.  The most expedient way to do that is to clobber one character
	 to null out the substring in-place.  Yecch. */
      while(remaining > 0) {
	char null_memory;
	char chunk_name[100];

	if(remaining > CHUNK_SIZE) {
	  null_memory = odlstring[CHUNK_SIZE];
	  odlstring[CHUNK_SIZE] = 0;
	}

	sprintf(chunk_name, "%s.%d", groupname, chunk_number);
	WriteDataset5(chunk_name, odlstring, (remaining > CHUNK_SIZE ? CHUNK_SIZE : remaining), gid);

	if(remaining > CHUNK_SIZE) {
	  odlstring[CHUNK_SIZE] = null_memory;
	}
	remaining -= CHUNK_SIZE;
	odlstring += CHUNK_SIZE;
      }
    
      /* We're done, close the group */
      if(H5Gclose(gid) < 0)
	return FAIL;

      /* Close and reopen hdfeos_handle */
      H5Fclose(hdf_handle);
      if(hdfeos_handle != he25_SWopen(hdfeos_file_name, H5F_ACC_RDWR))
	return FAIL;
    }
    TRACE((stderr, "ODL for group %s:\n%s\n", groupname, odlstring));
    return 0;
  }
}

/* Generic useful type conversion stuff */
/* maps platform-specific types to native common types that will be
   actually read in by HDF-EOS reader routines.  This will go away if
   the HDF-EOS developers take my suggestion and have the inquiry
   functions only emit "native" types...
*/

hid_t read_in_type(hid_t type) {
  switch(type) {
    /* the values we're going to map everything to - the "natives" */
  case HE5T_NATIVE_CHAR:  /* This one is the "STRING" value */
  case HE5T_NATIVE_SCHAR:
  case HE5T_NATIVE_UCHAR:
  case HE5T_NATIVE_INT:
  case HE5T_NATIVE_UINT:
  case HE5T_NATIVE_SHORT:
  case HE5T_NATIVE_USHORT:
  case HE5T_NATIVE_LONG:
  case HE5T_NATIVE_ULONG:
  case HE5T_NATIVE_LLONG:
  case HE5T_NATIVE_ULLONG:
  case HE5T_NATIVE_FLOAT:
  case HE5T_NATIVE_DOUBLE:
  case HE5T_NATIVE_LDOUBLE: return type;
    /* signed 8-bit */
  case  HE5T_NATIVE_INT8:
  case  HE5T_STD_I8BE:
  case  HE5T_STD_I8LE: return HE5T_NATIVE_SCHAR;
    /* unsigned 8-bit */
  case  HE5T_NATIVE_UINT8:
  case  HE5T_NATIVE_B8:
  case  HE5T_STD_U8BE:
  case  HE5T_STD_U8LE:
  case  HE5T_STD_B8BE:
  case  HE5T_STD_B8LE: return HE5T_NATIVE_UCHAR;
    /* signed 16-bit */
  case  HE5T_NATIVE_INT16:
  case  HE5T_STD_I16BE:
  case  HE5T_STD_I16LE: return HE5T_NATIVE_SHORT;
    /* unsigned 16-bit */
  case  HE5T_NATIVE_UINT16:
  case  HE5T_NATIVE_B16:
  case  HE5T_STD_U16BE:
  case  HE5T_STD_U16LE:
  case  HE5T_STD_B16BE:
  case  HE5T_STD_B16LE: return HE5T_NATIVE_USHORT;
    /* signed 32-bit */
  case  HE5T_NATIVE_INT32:
  case  HE5T_STD_I32BE:
  case  HE5T_STD_I32LE: return HE5T_NATIVE_INT;
    /* unsigned 32-bit */
  case  HE5T_NATIVE_UINT32:
  case  HE5T_NATIVE_B32:
  case  HE5T_STD_U32BE:
  case  HE5T_STD_U32LE:
  case  HE5T_STD_B32BE:
  case  HE5T_STD_B32LE: return HE5T_NATIVE_UINT;
    /* signed 64-bit */
  case  HE5T_NATIVE_INT64:
  case  HE5T_STD_I64BE:
  case  HE5T_STD_I64LE: return HE5T_NATIVE_LONG;
    /* unsigned 64-bit */
  case  HE5T_NATIVE_UINT64:
  case  HE5T_NATIVE_B64:
  case  HE5T_STD_U64BE:
  case  HE5T_STD_U64LE:
  case  HE5T_STD_B64BE:
  case  HE5T_STD_B64LE: return HE5T_NATIVE_ULONG;
    /* 32-bit float */
  case  HE5T_IEEE_F32BE:              
  case  HE5T_IEEE_F32LE: return HE5T_NATIVE_FLOAT;
    /* 64-bit float */
  case  HE5T_IEEE_F64BE:               
  case  HE5T_IEEE_F64LE: return  HE5T_NATIVE_DOUBLE;
    /* Weird stuff  and default */
  case  HE5T_NATIVE_HSIZE:
  case  HE5T_NATIVE_HERR:
  case  HE5T_NATIVE_HBOOL:
  default: return -1;
  }
}

/* and a little piece of junk that takes the "native" types and breaks out
   important stuff about them
*/
H5T_class_t type_info (hid_t type, int *size, int *signp)
{
  int hack;
  if(size == NULL) size = &hack;
  if(signp == NULL) signp = &hack;
  switch(type) {
    /* the values we're going to map everything to - the "natives" */
    /* one or both of these should map to H5T_STRING, I think... */
  case HE5T_NATIVE_CHAR:
    *size = 1; *signp = 0; return H5T_STRING;
  case HE5T_NATIVE_SCHAR:
    *size = 1; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_UCHAR:
    *size = 1; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_SHORT:
    *size = 2; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_USHORT:
    *size = 2; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_INT: 
    *size = 4; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_UINT:
    *size = 4; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_LONG:
    *size = 8; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_ULONG:
    *size = 8; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_LLONG:
    *size = 8; *signp = 1; return H5T_INTEGER;
  case HE5T_NATIVE_ULLONG:
    *size = 8; *signp = 0; return H5T_INTEGER;
  case HE5T_NATIVE_FLOAT:
    *size = 4; *signp = 1; return H5T_FLOAT;
  case HE5T_NATIVE_DOUBLE:
    *size = 8; *signp = 1; return H5T_FLOAT;
  case HE5T_NATIVE_LDOUBLE:
    *size = 8; *signp = 1; return H5T_FLOAT;
  default:
    TRACE((stderr, "Bogus value %d to type_info\n", type));
    return -1;
  }
}
/* One globally nasty little problem - H4 functions do dimension sizes, etc.
   as int32 (4 bytes), H5 functions do them as hsize_t (8 bytes in some places)

   This means we get to convert between them.  These helper functions
   do that, in-place (assuming an hsize_t field was filled in as if it were
   an int32).
*/

void int32_to_hsize_t(hsize_t *f, long fsize)
{
  if(f != NULL) {
    int32 *oldf = (int32 *)f;
    long i;

    for(i = fsize-1; i>=0; i--)
      f[i] = oldf[i];
  }
}

int32 *new_int32(const hsize_t *f, long fsize)
{
  int32 *result = calloc(fsize, sizeof(hsize_t));
  long i;

  for(i=0; i<fsize; i++)
    result[i] = f[i];
  return result;
}

/* Maps h4 type numbers to h5 type numbers - surprise! They're different! */
hid_t h4type_to_h5type(hid_t h4type)
{
  /* These are the only numbers that can be emitted as HDF-EOS4 types,
     according to the 2.9v1.00 sources.
  */
  switch(h4type) {
    /* I'm not 100% sure about these first two.  They appear in the HDF-EOS4
       sources as DFNT_UCHAR8 and DFNT_CHAR8, so I'm going to map them
       to the "string" data type.
    */
  case 3:
  case 4: return HE5T_NATIVE_CHAR;
  case 5: return HE5T_NATIVE_FLOAT;
  case 6: return HE5T_NATIVE_DOUBLE;
  case 20: return HE5T_NATIVE_SCHAR;
  case 21: return HE5T_NATIVE_UCHAR;
  case 22: return HE5T_NATIVE_SHORT;
  case 23: return HE5T_NATIVE_USHORT;
  case 24: return HE5T_NATIVE_INT;
  case 25: return HE5T_NATIVE_UINT;
  default:
    TRACE((stderr, "Unknown HDF-EOS4 data type! %ld", h4type));
    return 0;
  }
}

hid_t h5type_to_h4type(hid_t h5type) {
  switch(read_in_type(h5type)) {
    /* the values we're going to map everything to - the "natives" */
  case HE5T_NATIVE_CHAR:    return 4;
  case HE5T_NATIVE_SCHAR:   return 20;
  case HE5T_NATIVE_UCHAR:   return 21;
  case HE5T_NATIVE_UINT:     return 25;
  case HE5T_NATIVE_INT:    return 24;
  case HE5T_NATIVE_USHORT:   return 23;
  case HE5T_NATIVE_SHORT:  return 22;
  case HE5T_NATIVE_FLOAT:  return 5;
  case HE5T_NATIVE_DOUBLE: return 6;
  default:
    TRACE((stderr, "Unknown HDF-EOS5 data type! %ld\n", h5type));
    return 0;
  }
}



/* functions to map between H5T_NATIVE_mumble and HE5T_NATIVE_mumble spaces */
/* This one is useful for places where we want to send in HE5T but the
   underlying (broken, IMHO) API wants to see H5T */
hid_t hdfeos_native_to_hdf_native(hid_t heval) {
  switch(heval) {
  case HE5T_NATIVE_CHAR:    return H5T_NATIVE_CHAR;
  case HE5T_NATIVE_SCHAR:   return H5T_NATIVE_SCHAR;
  case HE5T_NATIVE_UCHAR:   return H5T_NATIVE_UCHAR;
  case HE5T_NATIVE_SHORT:   return H5T_NATIVE_SHORT; 
  case HE5T_NATIVE_USHORT:  return H5T_NATIVE_USHORT; 
  case HE5T_NATIVE_INT:     return H5T_NATIVE_INT;
  case HE5T_NATIVE_UINT:    return H5T_NATIVE_UINT;
  case HE5T_NATIVE_LLONG:   return H5T_NATIVE_LLONG;
  case HE5T_NATIVE_ULLONG:  return H5T_NATIVE_ULLONG;
  case HE5T_NATIVE_FLOAT:   return H5T_NATIVE_FLOAT;
  case HE5T_NATIVE_DOUBLE:  return H5T_NATIVE_DOUBLE;
  default:
    TRACE((stderr, "Unknown HDF-EOS5 data type! %ld\n", heval));
    return 0;
  }
}
/* Maybe I'm being dense here, but I can't see how to create this
   out of standard library functions... */
int count_char(char c, char const *s)
{
  int result = 0;
  char ctmp;
  while((ctmp = *s++) != 0)
    if(ctmp == c) result++;

  return result;
}
